
from dataclasses import dataclass,field

@dataclass
class CStruct:
    name: str
    """item name"""
    
    @classmethod
    def name_to_index(cls, name):
        raise NotImplementedError
    @classmethod
    def index_to_name(cls, index):
        raise NotImplementedError()    

    @property
    def index(self) -> int:
        """Index getter."""
        return self.name_to_index(self.name)
        
    @property
    def name(self) -> str:
        """Name getter."""
        return self._name
    @name.setter
    def name(self, name: str):
        """Name setter."""
        self._name = name
    
    def __repr__(self) -> str:
        return '<CStruct %s>' % self.name
    def __str__(self) -> str:
        """Converts the class to str, returns name."""
        return self.name
    def __int__(self) -> int:
        """Converts the class to int, returns index."""
        return self.index
    
    @classmethod
    def from_index(cls, index):
        """Constructs the class from index. Uses the reverse mapping onto name."""
        name = cls.index_to_name(index)
        return cls(name=name)



import logging
import numpy as np
#from pathlib import Path
import tempfile
import typing
from .dct_jpeg import DCTJPEG
from .spatial_jpeg import SpatialJPEG
from . import _jpeg
from ._colorspace import Colorspace

def read_dct(path: str):
    """Reads the DCT JPEG.
    
    :param path: Path to a source file in JPEG format.
    :type path: str
    :return: DCT JPEG object
    :rtype: :class:`DCTJPEG`
    :raises [IOError]: When source file does not exist
    
    :Example:
    
    >>> jpeg = jpeglib.read_jpeg_dct("input.jpeg")
    """
    # load file content
    with open(path, "rb") as f:
        content = f.read()
    # load info
    info = _jpeg.load_jpeg_info(path)
    # create jpeg
    return DCTJPEG(
        path                = path,
        content             = content,
        height              = info.height,
        width               = info.width,
        block_dims          = info.block_dims,
        samp_factor         = info.samp_factor,
        jpeg_color_space    = info.jpeg_color_space,
        num_components      = info.num_components,
        markers             = info.markers,
        Y                   = None,
        Cb                  = None,
        Cr                  = None,
        qt                  = None,
    )


    samp_factor: np.ndarray
    jpeg_color_space: Colorspace
    num_components: int

def read_spatial(
    path: str,
    out_color_space: str = None,
    dither_mode: str = None,
    dct_method: str = None,
    flags:list = []):
    """Decompresses the file into the spatial domain. 
    
    :param path: Path to a source file in JPEG format.
    :type path: str
    :param out_color_space: Output color space, must be accepted by :class:`_colorspace.Colorspace`. If not given, using the libjpeg default.
    :type out_color_space: str, optional
    :param dither_mode: Dither mode, must be accepted by :class:`_dithermode.Dithermode`. If not given, using the libjpeg default.
    :type dither_mode: str, optional
    :param dct_method: DCT method, must be accepted by :class:`_dctmethod.DCTMethod`. If not given, using the libjpeg default.
    :type dct_method: str, optional
    :param flags: Bool decompression parameters as str. If not given, using the libjpeg default. Read more at `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#flags>`_.
    :type flags: list, optional
    :return: Spatial JPEG object
    :rtype: :class:`SpatialJPEG`
    :raises [IOError]: When source file does not exist

    :Example:
    
    >>> jpeg = jpeglib.read_jpeg_spatial("input.jpeg")
    """
    # load file content
    with open(path, "rb") as f:
        content = f.read()
    # load info
    info = _jpeg.load_jpeg_info(path)
    # create jpeg
    return SpatialJPEG(
        path                = path,
        content             = content,
        height              = info.height,
        width               = info.width,
        block_dims          = info.block_dims,
        samp_factor         = info.samp_factor,
        jpeg_color_space    = info.jpeg_color_space,
        num_components      = info.num_components,
        markers             = info.markers,
        spatial             = None,
        color_space         = out_color_space,
        dither_mode         = dither_mode,
        dct_method          = dct_method,
        flags               = flags,
    )

def from_spatial(spatial: np.ndarray, in_color_space:typing.Union[str,Colorspace] = None):
    """Factory of spatial JPEG from spatial representation.
    
    Does not set the source path, you have to specify dstfile when writing.
        
    :param spatial: Spatial representation.
    :type spatial: np.ndarray
    :param in_color_space: Color space of the input. If not given, infered from the shape.
    :type in_color_space: str | Colorspace, optional
    
    :Example:
    
    >>> spatial = np.random.randint(0,255,(16,16,3),dtype=np.uint8)
    >>> jpeg = jpeglib.from_spatial(spatial)
    """
    # shape
    height,width,num_components = spatial.shape
    # parse colorspace
    if in_color_space is not None:
        try:
            in_color_space = Colorspace(in_color_space)
        except:
            in_color_space = in_color_space
    # infere colorspace
    if in_color_space is None:
        if num_components == 3:
            in_color_space = Colorspace('JCS_RGB')
        elif num_components == 1:
            in_color_space = Colorspace('JCS_GRAYSCALE')
        else:
            raise IOError('failed to infere colorspace')
    # create jpeg
    return SpatialJPEG(
        path                = None,
        content             = None,
        height              = height,
        width               = width,
        block_dims          = None,
        samp_factor         = None,
        jpeg_color_space    = None,
        num_components      = num_components,
        markers             = None,
        spatial             = spatial,
        color_space         = in_color_space,
        dither_mode         = None,
        dct_method          = None,
        flags               = [],
    )
    
    #with JPEG(srcfile) as im:
    #    Y,CbCr,qt = im.read_dct(*args, **kwargs)

# def write_dct(Y, CbCr=None, dstfile=None, qt=None, *args, **kwargs):
#     """Writes DCT coefficients to a file.
    
#     :param srcfile: Path to a source file in JPEG format. An implicit destination for writing.
#     :type srcfile: str
#     :param args: Positional arguments passed to JPEG.write_dct.
#     :param kwargs: Named arguments passed to JPEG.write_dct.
    
#     More information in documentation of JPEG.write_dct.
    
#     :Example:
    
#     >>> Y,CbCr,qt = jpeglib.read_dct("input.jpeg")
#     >>> jpeglib.write_dct(Y, CbCr, "output.jpeg", qt=qt)
#     """
#     assert(dstfile is not None)
#     with JPEG() as im:
#         im.write_dct(Y, CbCr, dstfile=dstfile, qt=qt, *args, **kwargs)

# def read_spatial(srcfile, *args, **kwargs):
#     """Decompresses the file into the spatial domain.
    
#     :param srcfile: Path to a source file in JPEG format. An implicit destination for writing.
#     :type srcfile: str
#     :param args: Positional arguments passed to JPEG.read_spatial.
#     :param kwargs: Named arguments passed to JPEG.read_spatial.
    
#     More information in documentation of JPEG.read_spatial.
    
#     :Example:
    
#     >>> x = jpeglib.read_spatial("input.jpeg")
#     >>> import matplotlib.pyplot as plt; plt.imshow(x)
#     """
#     with JPEG(srcfile) as im:
#         x = im.read_spatial(*args, **kwargs)
#         return x

# def write_spatial(data, dstfile, *args, **kwargs):
#     """Writes spatial image representation (i.e. RGB) to a file.
    
#     :param data: Numpy array with spatial data.
#     :type data: numpy.ndarray
#     :param dstfile: Destination file name. If not given, source file is overwritten.
#     :type dstfile: str, optional
#     :param args: Positional arguments passed to JPEG.write_spatial.
#     :param kwargs: Named arguments passed to JPEG.write_spatial.
    
#     More information in documentation of JPEG.read_spatial.
    
#     :Example:
    
#     >>> x = jpeglib.read_spatial("input.jpeg")
#     >>> jpeglib.write_spatial(x, "output.jpeg")
#     """
#     with JPEG() as im:
#         im.write_spatial(data, dstfile, *args, **kwargs)
    

    



import ctypes
from dataclasses import dataclass,fields
import numpy as np
import tempfile
from ._bind import CJpegLib
from . import _jpeg

@dataclass
class DCTJPEG(_jpeg.JPEG):
    """JPEG instance to work in DCT domain."""
    Y: np.ndarray
    """luminance"""
    Cb: np.ndarray
    """chrominance blue-difference"""
    Cr: np.ndarray
    """chrominance red-difference"""
    qt: np.ndarray
    """quantization"""
    
    def is_read(self) -> bool:
        has_Y = self.Y is not None
        has_qt = self.qt is not None
        has_CbCr = self.Cb is not None and self.Cr is not None
        has_no_CbCr = self.Cb is None and self.Cr is None
        return (
            (has_Y and has_qt and has_no_CbCr) # grayscale
            or (has_Y and has_qt and has_CbCr) # color
        )
    
    def _alloc_dct_component(self, i:int):
        return (((ctypes.c_short * 64) * self.width_in_blocks(i)) * self.height_in_blocks(i))()
    def read_dct(self):
        # write content into temporary file
        tmp = tempfile.NamedTemporaryFile(suffix='jpeg')
        tmp.write(self.content)
        # allocate DCT components
        Y = self._alloc_dct_component(0)
        Cb,Cr = None,None
        if self.has_chrominance(): # has chrominance
            Cb = self._alloc_dct_component(1)
            Cr = self._alloc_dct_component(2)
        qt = ((ctypes.c_short * 64) * 4)()
        # call
        CJpegLib.read_jpeg_dct(
            srcfile     = tmp.name,
            Y           = Y,
            Cb          = Cb,
            Cr          = Cr,
            qt          = qt
        )
        # close temporary file
        # process
        def process_component(comp):
            comp_np = np.ctypeslib.as_array(comp)
            return comp_np.reshape((*comp_np.shape[:-1],8,8))
        qt = process_component(qt)
        self.Y = process_component(Y)
        if self.has_chrominance():
            self.Cb = process_component(Cb)
            self.Cr = process_component(Cr)
        # crop
        self.qt = qt[:self.num_components]
        # return
        return self.Y,(self.Cb,self.Cr),self.qt
    
    def write_dct(self, path:str = None, quality:int=-1):
        """Writes DCT to JPEG.
        
        :param path: Destination file name. If not given, source file is overwritten.
        :type path: str, optional
        :param quality: Compression quality, between 0 and 100.
        :type quality: int, optional
        
        :Example:
        
        >>> jpeg = jpeglib.read_spatial("input.jpeg")
        >>> jpeg.write_spatial("output.jpeg", quality=92)
        """
        # write content into temporary file
        tmp = tempfile.NamedTemporaryFile(suffix='jpeg')
        tmp.write(self.content)
        # parameters
        dstfile = path if path is not None else self.path
        # convert dct
        def process_component(comp):
            comp = comp.reshape((*comp.shape[:-2],64))
            return np.ctypeslib.as_ctypes(comp.astype(np.int16))
        qt = process_component(self.qt)
        Y = process_component(self.Y)
        if self.has_chrominance():
            Cb = process_component(self.Cb)
            Cr = process_component(self.Cr)
        else: Cb,Cr = None,None
        # convert qt
        assert(quality in set(range(-1,101)))
        if quality != -1:
            qt = None
        # call
        CJpegLib.write_jpeg_dct(
            srcfile         = tmp.name,
            dstfile         = dstfile,
            Y               = Y,
            Cb              = Cb,
            Cr              = Cr,
            image_dims      = self.c_image_dims(),
            block_dims      = self.c_block_dims(),
            in_color_space  = self.jpeg_color_space.name,
            in_components   = self.num_components,
            qt              = qt,
            quality         = quality,
            num_markers     = self.num_markers(),
            marker_types    = self.c_marker_types(),
            marker_lengths  = self.c_marker_lengths(),
            markers         = self.c_markers(),
        )
    
    @property
    def Y(self) -> np.ndarray:
        """Luminance tensor getter."""
        if self._Y is None:
            self.read_dct()
        return self._Y
    @Y.setter
    def Y(self, Y: np.ndarray):
        """Luminance tensor setter."""
        self._Y = Y
    
    @property
    def Cb(self) -> np.ndarray:
        """Chrominance blue-difference tensor getter."""
        if self.has_chrominance() and self._Cb is None:
            self.read_dct()
        return self._Cb
    @Cb.setter
    def Cb(self, Cb: np.ndarray):
        """Chrominance blue-difference tensor setter."""
        self._Cb = Cb
    
    @property
    def Cr(self) -> np.ndarray:
        """Chrominance red-difference tensor getter."""
        if self.has_chrominance() and self._Cr is None:
            self.read_dct()
        return self._Cr
    @Cr.setter
    def Cr(self, Cr: np.ndarray):
        """Chrominance red-difference tensor setter."""
        self._Cr = Cr
    
    @property
    def qt(self) -> np.ndarray:
        """Quantization table getter."""
        if self._qt is None:
            self.read_dct()
        return self._qt
    @qt.setter
    def qt(self, qt: np.ndarray):
        """Quantization table setter."""
        self._qt = qt
    
    def free(self):
        """Free the allocated tensors."""
        del self._Y
        del self._Cb
        del self._Cr
        del self._qt
    
@dataclass
class DCTJPEGio(DCTJPEG):
    """Class for compatiblity with jpegio."""
    coef_arrays: list
    quant_tables: list
    def _convert_dct_jpegio(self, dct):
        return (dct
            .transpose((0,3,1,2))
            .reshape((dct.shape[0]*dct.shape[2], dct.shape[1]*dct.shape[3]))
        ).astype(np.int32)
    def _convert_jpegio_dct(self, dct):
        return (dct
            .reshape((-1,8,dct.shape[1]//8,8))
            .transpose((0,2,3,1))
        ).astype(np.int16)
    
    @property
    def coef_arrays(self):
        """Convertor of DCT coefficients to jpegio format."""
        if not self.is_read():
            self.read_dct()
        # collect dct
        self._coef_arrays = [
            self._convert_dct_jpegio(self.Y),
            self._convert_dct_jpegio(self.Cb),
            self._convert_dct_jpegio(self.Cr)
        ]
        return self._coef_arrays

    @coef_arrays.setter
    def coef_arrays(self, coef_arrays):
        self._coef_arrays = coef_arrays
    
    @property
    def quant_tables(self):
        """Convertor of quantization tables to jpegio format."""
        if not self.is_read():
            self.read_dct()
        self._quant_tables = [
            self.qt[0].astype(np.int32),
            self.qt[1].astype(np.int32)
        ]
        return self._quant_tables
    @quant_tables.setter
    def quant_tables(self, quant_tables):
        self._quant_tables = quant_tables
    
    def write(self, path:str = None, quality:int=-1):
        pass
        
        

def to_jpegio(jpeg: DCTJPEG):
    vals = {field.name: getattr(jpeg,field.name) for field in fields(jpeg)}
    return DCTJPEGio(**vals)




#!/usr/bin/env python3
"""Module with color space operations."""

from dataclasses import dataclass,field
from ._cstruct import CStruct

@dataclass
class Colorspace(CStruct):
    """Representation of color space.
    
    Allows mapping name-index and acquiral of number of channels.
    """

    _J_COLOR_SPACE = {
        "JCS_UNKNOWN":      0, # Unspecified color space
        "JCS_GRAYSCALE":    1, # Monochrome
        "JCS_RGB":          2, # Standard RGB
        "JCS_YCbCr":        3, # YCbCr or YUB, standard YCC
        "JCS_CMYK":         4, # CMYK
        "JCS_YCCK":         5, # YCbCrK
    }
    """List of color spaces. Taken from libjpeg internals."""
    
    @classmethod
    def J_COLOR_SPACE(cls) -> dict:
        """Get color space name to index mapping.
        
        :return: Dictionary, mapping color space name to index.
        :rtype: dict
        
        :Example:
        
        >>> name2index = Colorspace.J_COLOR_SPACE()
        >>> name2index['JCS_RGB'] # -> 2
        """
        return cls._J_COLOR_SPACE
    @classmethod
    def iJ_COLOR_SPACE(cls) -> dict:
        """Get index to color space name mapping.
        
        :return: Dictionary, mapping index to color space name.
        :rtype: dict
        
        :Example:
        
        >>> index2name = Colorspace.iJ_COLOR_SPACE()
        >>> index2name[1] # -> 'JCS_GRAYSCALE'
        """
        return {v: k for k, v in cls._J_COLOR_SPACE.items()}
    @classmethod
    def name_to_index(cls, name:str) -> int:
        """Map color space name onto index.
        
        :param name: Color space name.
        :type name: str
        :return: Index
        :rtype: int
        
        :Example:
        
        >>> Colorspace.name_to_index('JCS_GRAYSCALE') # -> 1
        >>> Colorspace.name_to_index('JCS_RGB') # -> 2
        """
        return cls.J_COLOR_SPACE()[name]
    @classmethod
    def index_to_name(cls, index:int):
        """Map index onto color space name.
        
        :param index: Index.
        :type name: int
        :return: Color space name
        :rtype: str
        
        :Example:
        
        >>> Colorspace.index_to_name(1) # -> 'JCS_GRAYSCALE')
        >>> Colorspace.index_to_name(2) # -> 'JCS_RGB')
        """
        return cls.iJ_COLOR_SPACE()[index]
    
    @property
    def channels(self) -> int:
        """Get channels of the colorspace.
        
        :return: Number of channels
        :rtype: int
        
        :Example:
        
        >>> colorspace = Colorspace("JCS_RGB")
        >>> colorspace # -> 3
        
        >>> colorspace = Colorspace("JCS_GRAYSCALE")
        >>> colorspace # -> 1
        """
        channel_no = {
            "JCS_GRAYSCALE": 1,
            "JCS_RGB":       3,
            "JCS_YCbCr":     3,
            "JCS_CMYK":      4,
            "JCS_YCCK":      4,
        }
        if self.name == "JCS_UNKNOWN":
            raise Exception("can't get number of channels for JCS_UNKNOWN colorspace")
        return channel_no[self.name]
    
    def __repr__(self):
        return '<Colorspace %s>' % self.name
    
    @classmethod
    def from_index(cls, index:int):
        """Construct from index.
        
        :param index: Color space index.
        :type index: int
        
        :Example:
        
        >>> colorspace = Colorspace.from_index(2)
        >>> colorspace.name # -> "JCS_RGB"
        """
        name = cls.iJ_COLOR_SPACE()[index]
        return cls(name=name)



import ctypes
from dataclasses import dataclass
import numpy as np
import tempfile
import typing
from ._bind import CJpegLib
from ._jpeg import JPEG
from ._colorspace import Colorspace
from ._dithermode import Dithermode
from ._dctmethod import DCTMethod

@dataclass
class SpatialJPEG(JPEG):
    """JPEG instance to work in spatial domain."""
    spatial: np.ndarray
    color_space: Colorspace
    dither_mode: Dithermode
    dct_method: DCTMethod
    flags: list
    
    def is_read(self) -> bool:
        return (rgb is not None) or (grayscale is not None)
    
    def _alloc_spatial(self, channels:int = None):
        if channels is None:
            channels = self.num_components
        return (((ctypes.c_ubyte * self.width) * self.height) * channels)()
    
    @classmethod
    def from_spatial(cls, spatial, ):
        """Factory of spatial JPEG from spatial representation.
        
        :param spatial: Spatial representation.
        :type spatial: np.ndarray

        :Example:
        
        >>> jpeg = jpeglib.read_spatial("input.jpeg")
        >>> jpeg.write_spatial("output.jpeg", qt=75)
        """
    
    def read_spatial(self):
        # write content into temporary file
        tmp = tempfile.NamedTemporaryFile(suffix='jpeg')
        tmp.write(self.content)
        
        # colorspace
        if self.color_space is None:
            self.color_space = self.jpeg_color_space
        # dither mode
        dither_mode = None
        if self.dither_mode is not None:
            dither_mode = self.dither_mode.index
        # dct method
        dct_method = None
        if dct_method is not None:
            dct_method = self.dct_method.index
            
        # allocate spatial
        spatial = self._alloc_spatial(self.color_space.channels)

        # call
        CJpegLib.read_jpeg_spatial(
            srcfile             = tmp.name,
            spatial             = spatial,
            colormap            = self.jpeg_color_space.index,
            in_colormap         = None, # support of color quantization
            out_color_space     = self.color_space.index,
            dither_mode         = dither_mode,
            dct_method          = dct_method,
            flags               = self.flags,
        )
        # process
        self.spatial = (
            np.ctypeslib.as_array(spatial)
            .astype(np.ubyte)
            .reshape(self.height, -1, self.color_space.channels)
        )
        return self.spatial

    def write_spatial(self, path:str = None, qt:typing.Union[int,np.ndarray] = None, dct_method:typing.Union[str,DCTMethod] = None, smoothing_factor:int = None, flags:list = []):
        """Writes a spatial image representation (i.e. RGB) to a file.
        
        :param path: Destination file name. If not given, source file is overwritten.
        :type path: str, optional
        :param qt: Compression quality, between 0 and 100 or a tensor with quantization tables. Defaultly -1 (default factor kept).
        :type qt: int | numpy.ndarray, optional
        :param dct_method: DCT method, must be accepted by :class:`_dctmethod.DCTMethod`. If not given, using the libjpeg default.
        :type dct_method: str | :class:`_dctmethod.DCTMethod`, optional
        :param smoothing_factor: Smoothing factor, between 0 and 100. Using default from libjpeg by default.
        :type smoothing_factor: int, optional
        :param flags: Bool compression parameters as str. If not given, using the libjpeg default. Read more at `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#flags>`_.
        :type flags: list, optional

        :Example:
        
        >>> jpeg = jpeglib.read_spatial("input.jpeg")
        >>> jpeg.write_spatial("output.jpeg", qt=75)
        """
        # colorspace
        if self.color_space is None:
            self.color_space = self.jpeg_color_space
        # path
        dstfile = path if path is not None else self.path
        if dstfile is None:
            raise IOError('no destination file specified')
        # dct method
        if dct_method is not None:
            # str
            try: dct_method = DCTMethod(dct_method).index
            # DCTMethod
            except: dct_method = dct_method.index
        # quality
        # use default of library
        if qt is None: quality,qt = -1,None
        else:
            # quality factor
            try: quality,qt = int(qt),None
            # quantization table
            except: quality,qt = -1,np.ctypeslib.as_ctypes(qt.astype(np.uint16))
        # process
        spatial = np.ctypeslib.as_ctypes(
            self.spatial.reshape(
                self.color_space.channels,
                self.height,
                self.width,
            )
        )
        # call
        CJpegLib.write_jpeg_spatial(
            dstfile             = dstfile,
            spatial             = spatial,
            image_dims          = self.c_image_dims(),
            in_color_space      = self.color_space.index,
            in_components       = self.color_space.channels,
            dct_method          = dct_method,
            samp_factor         = self.c_samp_factor(),
            qt                  = qt,
            quality             = quality,
            smoothing_factor    = smoothing_factor,
            num_markers         = self.num_markers(),
            marker_types        = self.c_marker_types(),
            marker_lengths      = self.c_marker_lengths(),
            markers             = self.c_markers(),
            flags               = flags,
        )
    
    @property
    def spatial(self) -> np.ndarray:
        if self._spatial is None:
            self.read_spatial()
        return self._spatial
    @spatial.setter
    def spatial(self, spatial: np.ndarray):
        self._spatial = spatial
    
    @property
    def color_space(self) -> np.ndarray:
        return self._color_space
    @color_space.setter
    def color_space(self, color_space: Colorspace):
        self._color_space = color_space
    
    @property
    def dither_mode(self) -> np.ndarray:
        return self._dither_mode
    @dither_mode.setter
    def dither_mode(self, dither_mode: Dithermode):
        self._dither_mode = dither_mode
    
    @property
    def dct_method(self) -> np.ndarray:
        return self._dct_method
    @dct_method.setter
    def dct_method(self, dct_method: DCTMethod):
        self._dct_method = dct_method
    
    @property
    def flags(self) -> list:
        return self._flags
    @flags.setter
    def flags(self, flags: list):
        self._flags = flags
    
    def free(self):
        """Free the allocated tensors."""
        del self._spatial
    


"""
Python envelope for the popular C library libjpeg for handling JPEG files.

It offers full control over compression and decompression and exposes DCT coefficients and quantization tables.
"""

# functions
from .functional import *

# jpeg objects
from .dct_jpeg import *
from .spatial_jpeg import *

# cstructs
from ._colorspace import *
from ._dctmethod import *
from ._dithermode import *
from ._marker import *

# libjpeg versions
from .version import *

# for unit tests
from ._timer import Timer

# package version
import pkg_resources
try:
    __version__ = pkg_resources.get_distribution("jpeglib").version
except:
    __version__ = None

# set default version
version.set('6b')



import ctypes
import os
import pathlib
import re

from . import cjpeglib

class CJpegLib:

    @classmethod
    def jpeg_lib_version(cls):
        return cls.get().jpeg_lib_version()

    @classmethod
    def read_jpeg_info(cls,
        srcfile:str, block_dims, image_dims, num_components,
        samp_factor, jpeg_color_space, marker_lengths, marker_types):
        status = cls.get().read_jpeg_info(cls.cstr(srcfile), block_dims, image_dims, num_components,
                                          samp_factor, jpeg_color_space, marker_lengths, marker_types)
        if status == 0: raise IOError(f"reading info of {srcfile} failed")
    
    @classmethod
    def read_jpeg_markers(cls, srcfile:str, markers):
        status = cls.get().read_jpeg_markers(cls.cstr(srcfile), markers)
        if status == 0: raise IOError(f"reading markers of {srcfile} failed")
        
    @classmethod
    def read_jpeg_dct(cls, srcfile:str, Y, Cb, Cr, qt):
        status = cls.get().read_jpeg_dct(cls.cstr(srcfile), Y, Cb, Cr, qt)
        if status == 0: raise IOError(f"reading of {srcfile} DCT failed")

    @classmethod
    def write_jpeg_dct(cls,
        srcfile:str, dstfile:str, Y, Cb, Cr,
        image_dims, block_dims, in_color_space, in_components,
        qt, quality, num_markers, marker_types, marker_lengths, markers):
        status = cls.get().write_jpeg_dct(cls.cstr(srcfile), cls.cstr(dstfile), Y, Cb, Cr,
                                          image_dims, block_dims, in_color_space, in_components,
                                          qt, quality, num_markers, marker_types, marker_lengths, markers)
        if status == 0: raise IOError(f"writing DCT to {dstfile} failed")
    @classmethod
    def print_jpeg_params(cls, srcfile:str):
        status = cls.get().print_jpeg_params(cls.cstr(srcfile))
        if status == 0: raise IOError(f"reading of {srcfile} failed")

    @classmethod
    def read_jpeg_spatial(cls,
        srcfile:str, spatial, colormap, in_colormap, out_color_space,
        dither_mode, dct_method, flags):
        status = cls.get().read_jpeg_spatial(cls.cstr(srcfile), spatial, colormap, in_colormap,
                                             out_color_space, dither_mode, dct_method, cls.flags_to_mask(flags))
        if status == 0: raise IOError(f"reading of {srcfile} spatial failed")
    
    @classmethod
    def write_jpeg_spatial(cls,
        dstfile:str, spatial, image_dims, in_color_space, in_components,
        dct_method, samp_factor, qt, quality, smoothing_factor,
        num_markers:int, marker_types, marker_lengths, markers,
        flags
    ):
        status = cls.get().write_jpeg_spatial(cls.cstr(dstfile), spatial, image_dims, in_color_space, in_components,
                                              dct_method, samp_factor, qt, cls.factor(quality), cls.factor(smoothing_factor),
                                              num_markers, marker_types, marker_lengths, markers,
                                              cls.flags_to_mask(flags))
        if status == 0: raise IOError(f"writing RGB to {dstfile} failed")
        
    MASKS = {
        "DO_FANCY_SAMPLING": (0b1 << 0), "DO_FANCY_UPSAMPLING": (0b1 << 0), "DO_FANCY_DOWNSAMPLING": (0b1 << 0),
        "DO_BLOCK_SMOOTHING": (0b1 << 2),
        "TWO_PASS_QUANTIZE": (0b1 << 4),
        "ENABLE_1PASS_QUANT": (0b1 << 6),
        "ENABLE_EXTERNAL_QUANT": (0b1 << 8),
        "ENABLE_2PASS_QUANT": (0b1 << 10),
        "OPTIMIZE_CODING": (0b1 << 12),
        "PROGRESSIVE_MODE": (0b1 << 14),
        "QUANTIZE_COLORS": (0b1 << 16),
        "ARITH_CODE": (0b1 << 18),
        "WRITE_JFIF_HEADER": (0b1 << 20),
        "WRITE_ADOBE_MARKER": (0b1 << 22),
        "CCIR601_SAMPLING": (0b1 << 24),
        "FORCE_BASELINE": (0b1 << 26),
    }

    @classmethod
    def flags_to_mask(cls, flags):
        mask = 0xFFFFFFFF
        if flags is None: return mask
        for flag in flags:
            # print(f"flag {flag}:")
            # parse sign
            sign = '-' if flag[0] == '-' else '+'
            if not flag[0].isalpha(): flag = flag[1:]
            # get flags
            flagbit = cls.MASKS[flag.upper()]
            defbit = cls.MASKS[flag.upper()] << 1
            # map
            mask ^= defbit # reset default value
            if sign == '-':
                mask ^= (flagbit) # erase bit

        return ctypes.c_ulonglong(mask)
    @classmethod
    def factor(cls, factor):
        if factor is None:
            factor = 0
        return ctypes.c_short(factor)
    
    _lib = None
    version = None
    @classmethod
    def get(cls):
        # connect to library
        if cls._lib is None:
            cls._lib = cls._bind_lib()
        # return library
        return cls._lib
    
    @classmethod
    def set_version(cls, version):
        cls._lib = cls._bind_lib(version=version)
    @classmethod
    def get_version(cls):
        return cls.version
    @classmethod
    def _versions(cls):
        so_files = [f for f in os.listdir(list(cjpeglib.__path__)[0]) if re.fullmatch(f'cjpeglib_.*\..*\.so', f)]
        return so_files
    @classmethod
    def versions(cls):
        vs = [re.search(f'cjpeglib_[^.]*\..*\.so', f) for f in cls._versions()]
        vs = [v[0] for v in vs if v]
        return vs
    @classmethod
    def _bind_lib(cls, version='6b'):
        # path of the library
        so_files = [f for f in cls._versions() if re.fullmatch(f'cjpeglib_{version}\..*\.so', f)]
        try:
            so_file = so_files[0]
        except:
            raise Exception(f"dynamic library not found")
        libname = pathlib.Path(list(cjpeglib.__path__)[0]) / so_file
        # connect
        cjpeglib_dylib = ctypes.CDLL(libname)
        cls.version = version
        return cjpeglib_dylib
    
    @staticmethod
    def cstr(s):
        if s is None: return None
        return ctypes.c_char_p(s.encode('utf-8'))
        



from dataclasses import dataclass,field
from ._cstruct import CStruct

@dataclass
class Marker(CStruct):
    
    _J_MARKER_CODE = {
        "JPEG_RST0":    0xD0, #
        "JPEG_RST1":    0xD0+1, #
        "JPEG_RST2":    0xD0+2, #
        "JPEG_RST3":    0xD0+3, #
        "JPEG_RST4":    0xD0+4, #
        "JPEG_RST5":    0xD0+5, #
        "JPEG_RST6":    0xD0+6, #
        "JPEG_RST7":    0xD0+7, #
        "JPEG_RST8":    0xD0+8, #
        "JPEG_EOI":     0xD9, #
        "JPEG_APP0":    0xE0, #
        "JPEG_APP1":    0xE0+1, #
        "JPEG_APP2":    0xE0+2, #
        "JPEG_APP3":    0xE0+3, #
        "JPEG_APP4":    0xE0+4, #
        "JPEG_APP5":    0xE0+5, #
        "JPEG_APP6":    0xE0+6, #
        "JPEG_APP7":    0xE0+7, #
        "JPEG_APP8":    0xE0+8, #
        "JPEG_APP9":    0xE0+9, #
        "JPEG_APP10":   0xE0+10, #
        "JPEG_APP11":   0xE0+11, #
        "JPEG_APP12":   0xE0+12, #
        "JPEG_APP13":   0xE0+13, #
        "JPEG_APP14":   0xE0+14, #
        "JPEG_APP15":   0xE0+15, #
        "JPEG_COM":     0xFE,
    }
    length:int
    content:bytes
    
    @classmethod
    def J_MARKER_CODE(cls):
        return cls._J_MARKER_CODE
    @classmethod
    def iJ_MARKER_CODE(cls):
        return {v: k for k, v in cls._J_MARKER_CODE.items()}
    
    @classmethod
    def name_to_index(cls, name):
        return cls.J_MARKER_CODE()[name]
    @classmethod
    def index_to_name(cls, index):
        return cls.iJ_MARKER_CODE()[index]

    def __repr__(self):
        data_size = 0
        if self.content is not None:
            data_size = len(self.content)
        return '<Marker %s N=%s>' % (self.name, data_size)
    def __len__(self):
        return self.length

    @property
    def length(self) -> int:
        return self._length
    @length.setter
    def length(self, length: int):
        self._length = length
        
    @property
    def content(self) -> bytes:
        return self._content
    @content.setter
    def content(self, content: bytes):
        self._content = content
    
    @classmethod
    def from_index(cls, index:int, length:int, content:bytes):
        name = cls.index_to_name(index)
        return cls(name=name, length=length, content=content)



import ctypes
from collections import namedtuple
from dataclasses import dataclass
import logging
import numpy as np
from ._bind import CJpegLib
from ._colorspace import Colorspace
from ._marker import Marker

@dataclass
class JPEG:
    """JPEG abstract class."""
    path: str
    content: bytes
    height: int
    width: int
    block_dims: np.ndarray
    samp_factor: np.ndarray
    jpeg_color_space: Colorspace
    num_components: int
    markers: list 
    
    def height_in_blocks(self, component):
        return self.block_dims[component][0]
    def width_in_blocks(self, component):
        return self.block_dims[component][1]
    def has_chrominance(self):
        return self.num_components > 1
    def num_markers(self) -> int:
        if self.markers is None:
            return 0
        return len(self.markers)
    
    def c_image_dims(self):
        return (ctypes.c_int * 2)(self.height,self.width)
    def c_block_dims(self):
        return (ctypes.c_int * 6)(
            self.block_dims[0][0], self.block_dims[0][1],
            self.block_dims[1][0], self.block_dims[1][1],
            self.block_dims[2][0], self.block_dims[2][1],)
    def c_samp_factor(self):
        if self.samp_factor is None:
            return self.samp_factor
        samp_factor = np.array(self.samp_factor, dtype=np.int32)
        return np.ctypeslib.as_ctypes(samp_factor)
    def c_marker_types(self):
        if self.markers is None:
            return None
        marker_types = [marker.index for marker in self.markers]
        return (ctypes.c_int32*self.num_markers())(*marker_types)
    def c_marker_lengths(self):
        if self.markers is None:
            return None
        marker_lengths = [marker.length for marker in self.markers]
        return (ctypes.c_int32*self.num_markers())(*marker_lengths)
    def c_markers(self):
        if self.markers is None:
            return None
        marker_lengths = [marker.length for marker in self.markers]
        marker_contents = []
        for marker in self.markers:
            marker_contents += [i for i in marker.content]
        return (ctypes.c_ubyte*np.sum(marker_lengths))(*marker_contents)
        
    def free(self):
        """Free the allocated tensors."""
        raise NotImplementedError


def load_jpeg_info(path: str):
    """"""
    # allocate
    _block_dims = (ctypes.c_int*6)()
    _image_dims = (ctypes.c_int*2)()
    _num_components = (ctypes.c_int*1)()
    _samp_factor = (ctypes.c_int*6)()
    _jpeg_color_space = (ctypes.c_int*1)()
    _marker_lengths = (ctypes.c_int*20)()
    _marker_types = (ctypes.c_uint32*20)()
    # call
    CJpegLib.read_jpeg_info(
        srcfile             = str(path),
        block_dims          = _block_dims,
        image_dims          = _image_dims,
        num_components      = _num_components,
        samp_factor         = _samp_factor,
        jpeg_color_space    = _jpeg_color_space,
        marker_lengths      = _marker_lengths,
        marker_types        = _marker_types,
    )
    # process
    num_components = _num_components[0] # number of components in JPEG
    block_dims = (
        np.array([_block_dims[i] for i in range(2*num_components)], int)
        .reshape(num_components, 2))
    samp_factor = (
        np.array([_samp_factor[i] for i in range(2*num_components)], int)
        .reshape(num_components, 2))
    markers = []
    for i in range(20):
        # marker length
        marker_length = _marker_lengths[i]
        if marker_length == 0:
            break
        # marker type
        marker_type = _marker_types[i]
        # create marker
        marker = Marker.from_index(
            index=marker_type, 
            length=marker_length, 
            content=None,
        )
        markers.append(marker)
    marker_lengths = np.array([marker.length for marker in markers], dtype=np.int32)
    num_markers = marker_lengths.shape[0]
    
    # allocate
    _markers = (ctypes.c_ubyte * np.sum(marker_lengths))()
    # call
    CJpegLib.read_jpeg_markers(
        srcfile             = str(path),
        markers             = _markers,
    )
    # process
    cumlens = np.cumsum([0] + marker_lengths.tolist())
    for i in range(num_markers):
        markers[i].content = bytes(_markers[cumlens[i]:cumlens[i+1]])
    # create jpeg
    return JPEG(
        path                = path,
        height              = _image_dims[0],
        width               = _image_dims[1],
        block_dims          = block_dims,
        samp_factor         = samp_factor,
        num_components      = num_components,
        jpeg_color_space    = Colorspace.from_index(_jpeg_color_space[0]),
        content             = None,
        markers             = markers,
    )




import math

def GCD(*args):
    """Computes greatest common divider (gcd) of the positional arguments."""
    # replace 0s with 1s
    args = [a if a != 0 else 1 for a in args]
    # all divisors
    divisors = [i for i in range(1,min(args)+1) if all([x % i == 0 for x in args])]
    # choose the maximal
    gcd = max(divisors)
    return gcd

def Jab_to_factors(Jab):
    """Converts J:a:b notation to the 6 factors, used by libjpeg.
    
    Read more about the notations in `Tuomas Siipola's article <https://zpl.fi/chroma-subsampling-and-jpeg-sampling-factors/>`_
    or `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#jpeg-sampling-factor>`_.
    
    :param Jab: List of three numbers, J:a:b notation of subsampling.
    :type dstfile: list
    :return: Subsampling factors, as list of three lists, each containing 2 numbers.
    :rtype: list

    :Example:

    >>> Jab_to_factors([4,2,0]) # -> [[2,2],[1,1],[1,1]]
    """
    # parse input
    J,a,b = Jab
    assert(J in {1,2,3,4})
    if J == 4: assert(a in {4,2,1,0})
    elif J == 3: assert(a in {3,1,0})
    else: assert(a in {J,1,0})
    assert(b in {a,0})
    # chroma dimensions
    Y = [J,2]
    Cb = Cr = [a,(a == b) + 1]
    # normalize by GCD
    gcd0 = GCD(Y[0], Cb[0], Cr[0])
    gcd1 = GCD(Y[1], Cb[1], Cr[1])
    factors = [
        [int(Y[0] / gcd0), int(Y[1] / gcd1)],
        [int(Cb[0] / gcd0), int(Cb[1] / gcd1)],
        [int(Cr[0] / gcd0), int(Cr[1] / gcd1)],
    ]
    return factors
    
if __name__ == '__main__':
    Jab_to_factors([4,4,4])
    Jab_to_factors([4,4,0])
    Jab_to_factors([4,2,2])
    Jab_to_factors([4,2,0])
    Jab_to_factors([4,1,1])
    Jab_to_factors([4,1,0])
    Jab_to_factors([3,3,3])
    Jab_to_factors([3,3,0])
    Jab_to_factors([3,1,1])
    Jab_to_factors([3,1,0])
    



from dataclasses import dataclass,field
from ._cstruct import CStruct

@dataclass
class Dithermode(CStruct):
    
    _J_DITHER_MODE = {
        "JDITHER_NONE":     0, #
        "JDITHER_ORDERED":  1, #
        "JDITHER_FS":       2, #
    }
    @classmethod
    def J_DITHER_MODE(cls):
        return cls._J_DITHER_MODE
    @classmethod
    def iJ_DITHER_MODE(cls):
        return {v: k for k, v in cls._J_DITHER_MODE.items()}
    
    @classmethod
    def name_to_index(cls, name):
        return cls.J_DITHER_MODE()[name]
    @classmethod
    def index_to_name(cls, index):
        return cls.iJ_DITHER_MODE()[index]

    def __repr__(self):
        return '<Dithermode %s>' % self.name



from dataclasses import dataclass,field
from ._cstruct import CStruct

@dataclass
class DCTMethod(CStruct):
    
    _J_DCT_METHOD = {
        "JDCT_ISLOW":   0, #
        "JDCT_IFAST":   1, #
        "JDCT_FLOAT":   2, #
    }
    @classmethod
    def J_DCT_METHOD(cls):
        return cls._J_DCT_METHOD
    @classmethod
    def iJ_DCT_METHOD(cls):
        return {v: k for k, v in cls._J_DCT_METHOD.items()}
    
    @classmethod
    def name_to_index(cls, name):
        return cls.J_DCT_METHOD()[name]
    @classmethod
    def index_to_name(cls, index):
        return cls.iJ_DCT_METHOD()[index]

    def __repr__(self):
        return '<DCTMethod %s>' % self.name



from ._bind import CJpegLib

class version:
    """Class grouping functions for controlling libjpeg method."""
    @staticmethod
    def set(version):
        """Sets the version of libjpeg to use. Loads the library.
        
        :param version: libjpeg version, one of 6b, 8d, 9d, turbo210.
        :type version: str
        :raises [NotImplementedError]: unsupported libjpeg version

        :Example:

        >>> import jpeglib
        >>> jpeglib.version.set('8d')
        """
        if version in {'6','6b'}:
            CJpegLib.set_version(version='6b')
        elif version in {'7','8','8a','8b','8c','8d','9','9a','9b','9c','9d','9e'}:
            CJpegLib.set_version(version=version)
        elif version in {'turbo','turbo2.1.0','turbo2.1','turbo210','turbo21'}:
            CJpegLib.set_version(version='turbo210')
        elif version in {'mozjpeg4.0.3','mozjpeg4','mozjpeg403','mozjpeg'}:
            CJpegLib.set_version(version='mozjpeg403')
        else:
            raise NotImplementedError(f'Unsupported libjpeg version')
    @staticmethod
    def get():
        """Gets the currently used version of libjpeg. 
        
        :return: libjpeg version or None if not been loaded yet.
        :rtype: str, None

        :Example:

        >>> import jpeglib
        >>> jpeglib.version.set('6b')
        >>> jpeglib.version.get()
        '6b'
        """
        return CJpegLib.get_version()
    @staticmethod
    def _jpeg_lib_version():
        """Returns value of jpeg_lib_version macro."""
        return CJpegLib.jpeg_lib_version()
    @staticmethod
    def _get_lib():
        """Low-level getter of the dynamic library.
        
        :return: Dynamic library object or None if not loaded yet.
        :rtype: ctypes.CDLL, None
        """
        return CJpegLib.get()
    @staticmethod
    def versions():
        """Lists present DLLs of versions."""
        return CJpegLib.versions()
    def __init__(self, version):
        """Constructor, used in with statement.
        
        :param version: Version to set inside with block.
        :type version: str
        """
        self.next = version
    def __enter__(self):
        """Sets new version in a block.

        :Example:

        >>> jpeglib.version.set('6b')
        >>> # working with 6b
        >>> # [...]
        >>> with jpeglib.version('8d'):
        >>>     # working with 8d
        >>>     # [...]
        >>>     pass
        """
        self.prev = self.get()
        self.set(self.next)
    def __exit__(self, *args, **kw):
        """Recovers a previous version, when exiting `with` block.
        
        :Example:

        >>> # working with 6b
        >>> # [...]
        >>> with jpeglib.version('8d'):
        >>>     # working with 8d
        >>>     # [...]
        >>>     pass
        >>> # working with 6b (again)
        >>> # [...]
        """
        self.set(self.prev)

__all__ = ['version']


"""
Module containing conversion of C-structures into numpy and back and memory management.
"""

import ctypes
import logging
import math
import numpy as np
import tempfile
import warnings

from ._bind import CJpegLib
from ._timer import Timer
from ._notations import Jab_to_factors

class JPEG:
    cjpeglib = CJpegLib
    def __init__(self, srcfile=None):
        """Constructs the :class:`JPEG` object.
        
        :param srcfile: Path to a source file in JPEG format. An implicit destination for writing.
        :type srcfile: str
        :raises [IOError]: When file does not exist or can't be read.
        
        :Example:

        >>> im = jpeglib.JPEG("input.jpeg")
        """
        # set filename
        self.srcfile = srcfile
        # allocate resources
        self._dct_dims = (ctypes.c_int*6)()
        self._dims = (ctypes.c_int*3)()
        self._num_components = (ctypes.c_int*1)()
        self._im_qt = ((ctypes.c_short*64)*(4))()
        self._samp_factor = None # self._parse_samp_factor()
        self.color_space = 'JCS_RGB'
        #self.color_space = [k for k,v in self.J_COLOR_SPACE.items() if v[0] == self._color_space[0]][0]
        # get image info
        if srcfile is not None:
            try:
                self._read_info()
                self._allocate() # allocate
            except:
                self._im_spatial = None
                self._im_colormap = None
                self._im_dct = None
        else:
            self._im_spatial = None
            self._im_colormap = None
            self._im_dct = None

    def read_dct(self, quantized=False):
        """Reads the DCT coefficients and quantization tables of the source file.

        In the return values (when format is 'block'), Y is DCT luminance tensor of shape (1,W/8,H/8,8,8),
        CbCr is DCT chrominance tensor of shape (2,W/8/sw,H/8/sh,8,8), both not-quantized by default,
        where sw, sh are corresponsing chrominance scaling factors.
        qt is a tensor with quantization tables of shape (2,8,8).
        
        When format is 'jpegio', DCT tensors have shapes (W,H) and (W/sw, H/sh).
        Check the examples below.

        :param quantized: Indicates whether the output DCT coefficients are quantized or not, False by default.
        :type quantized: bool, optional
        :param format: Format of output, one of 'block', 'jpegio'.
        :type format: str, optional
        :return: lumo DCT Y, chroma DCT CbCr, quantization table qt
        :rtype: tuple
        :raises [IOError]: When source file was not given in constructor.

        :Example:

        >>> im = jpeglib.JPEG("input.jpeg")
        >>> Y,CbCr,qt = im.read_dct()

        Get DCT coefficients in shape (W/8, H/8, 8, 8) with

        >>> Y[0]    # Y
        >>> CbCr[0] # Cb
        >>> CbCr[1] # Cr

        Get quantization tables in shape (8,8) with

        >>> qt[0] # Y
        >>> qt[1] # Cb
        >>> qt[1] # Cr
        
        Get DCT coefficients in shape (W, H) with
        
        >>> Y,CbCr,qt = im.read_dct(format='jpegio')
        """
        # execute
        Y,CbCr,qt = self._read_dct(self.srcfile)
        # quantization
        if quantized: Y,CbCr = Y*qt[0],CbCr*qt[1]
        # format
        if format == 'block': pass
        elif format == 'jpegio':
            Y = (Y
                .transpose((0,2,4,1,3))
                .reshape((Y.shape[2]*Y.shape[4], Y.shape[1]*Y.shape[3]))
            )
            CbCr = (CbCr
                    .transpose((0,2,4,1,3))
                    .reshape((Y.shape[2]*Y.shape[4], Y.shape[1]*Y.shape[3]))
            )
        # result
        return Y,CbCr,qt

    def write_dct(self, Y, CbCr=None, dstfile=None, qt=None, quantized=False, format='block', in_color_space=None, samp_factor=None):
        """Writes DCT coefficients to a file.
        
        The shape and arrangement of the parameters Y and CbCr is the same
        as what is returned from :meth:`jpeg.JPEG.read_dct`.
        Y is DCT luminance tensor of shape (1,W/8,H/8,8,8), CbCr is DCT chrominance tensor of shape (2,W/8,H/8,8,8),
        both unquantized by default. qt is either a tensor with quantization tables of shape (2,8,8) or a quality scalar integer.

        :param Y: Lumo DCT.
        :type Y: numpy.ndarray
        :param CbCr: Chrominance DCT.
        :type CbCr: numpy.ndarray, optional
        :param dstfile: Destination file. If not given, source file is overwritten.
        :type dstfile: str, optional
        :param qt: Quantization tables or quality.
        :type qt: numpy.ndarray | int, optional
        :param quantized: Indicates whether the input DCT coefficients are quantized or unquantized, False by default.
        :type quantized: bool, optional 
        :param in_color_space: Input color space. Must be key of :class:`jpeg.JPEG.J_COLOR_SPACE`. According to source by default.
        :type in_color_space: str, optional
        :param samp_factor: Sampling factor. None, or list of 3 tuples of 2 ints. According to source by default. Read more at `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#jpeg-sampling-factor>`_.
        :type samp_factor: list, optional

        :Example:

        >>> im = jpeglib.JPEG("input.jpeg")
        >>> Y,CbCr,qt = im.read_dct()
        >>> # work with DCT coefficients
        >>> # ...
        >>> im.write_dct("output.jpeg", Y, CbCr, qt)

        You can specify the quality scalar such as

        >>> im.write_dct("output.jpeg", Y, CbCr, 75)

        If neither quality nor quantization tables are specified,
        library uses the quantization table from the source.
        """
        # choose destination
        if dstfile is None:
            if self.srcfile is None:
                raise Exception("destination not specified")
            dstfile = self.srcfile
        # execute
        self._write_dct(dstfile, Y, CbCr, qt, quantized, in_color_space, samp_factor)
        # free buffers
        self._im_dct = None
        self._im_spatial = None

    # https://bitmiracle.github.io/libjpeg.net/help/api/BitMiracle.LibJpeg.Classic.J_COLOR_SPACE.html
    J_COLOR_SPACE = {
        None: (-1,-1), # default
        "JCS_UNKNOWN":(0,0), # Unspecified color space
        "JCS_GRAYSCALE":(1,1), # Monochrome
        "JCS_RGB":(2,3), # Standard RGB
        "JCS_YCbCr":(3,3), # YCbCr or YUB, standard YCC
        "JCS_CMYK":(4,4), # CMYK
        "JCS_YCCK":(5,4) # YCbCrK
    }
    J_DITHER_MODE = {
        None: -1,
        "JDITHER_NONE": 0,
        "JDITHER_ORDERED": 1,
        "JDITHER_FS": 2
    }
    J_DCT_METHOD = {
        None: -1,
        "JDCT_ISLOW": 0, # slow but accurate integer algorithm
        "JDCT_IFAST": 1, # faster, less accurate integer method
        "JDCT_FLOAT": 2, # floating-point method
    }

    def read_spatial(self, out_color_space=None, dither_mode=None, dct_method=None, colormap=None, flags=[]):
        """Decompresses the file into the spatial domain. 
        
        :param out_color_space: Output color space. Must be key of J_COLOR_SPACE.
        :type out_color_space: str, optional
        :param dither_mode: Dither mode. Must be key of :class:`jpeg.JPEG.J_DITHER_MODE`. Using default from libjpeg by default.
        :type dither_mode: str, optional
        :param dct_method: DCT method. Must be key of :class:`jpeg.JPEG.J_DCT_METHOD`. Using default from libjpeg by default.
        :type dct_method: str, optional
        :param colormap: Colormap to use for color quantization.
        :type colormap: np.array, optional
        :param flags: Bool decompression parameters as str to set to be true. Using default from libjpeg by default.
        :type flags: list, optional
        :return: Spatial representation of the source image.
        :rtype: numpy.ndarray

        :Example:

        >>> im = jpeglib.JPEG("input.jpeg")
        >>> rgb = im.read_spatial(out_color_space="JCS_RGB")
        """
        # execute
        spatial = self._read_spatial(self.srcfile, out_color_space, dither_mode, dct_method, colormap, flags)
        # free buffers
        self._im_dct = None
        # result
        return spatial

    def write_spatial(self, data, dstfile=None, in_color_space=None, dct_method="JDCT_ISLOW",
                      samp_factor=None, qt=-1, smoothing_factor=None, flags=[]):
        """Writes spatial image representation (i.e. RGB) to a file.
        
        :param data: Numpy array with spatial data.
        :type data: numpy.ndarray
        :param dstfile: Destination file name. If not given, source file is overwritten.
        :type dstfile: str, optional
        :param in_color_space: Input color space. Must be key of :class:`jpeg.JPEG.J_COLOR_SPACE`. According to source by default.
        :type in_color_space: str, optional
        :param dct_method: DCT method. Must be key of :class:`jpeg.JPEG.J_DCT_METHOD`. Using default from libjpeg by default.
        :type dct_method: str, optional
        :param samp_factor: Sampling factor. None, or list of 3 tuples of 2 ints. According to source by default. Read more at `glossary <https://jpeglib.readthedocs.io/en/latest/glossary.html#jpeg-sampling-factor>`_.
        :type samp_factor: list, optional
        :param qt: Compression quality, between 0 and 100 or a tensor with quantization tables. Defaultly -1 (default factor kept).
        :type qt: int | numpy.ndarray, optional
        :param smoothing_factor: Smoothing factor, between 0 and 100. Using default from libjpeg by default.
        :type smoothing_factor: int, optional
        :param flags: Bool decompression parameters as str to set to be true. Using default from libjpeg by default.
        :type flags: list, optional

        :Example:

        >>> im = jpeglib.JPEG("input.jpeg")
        >>> grayscale = im.read_spatial(out_color_space="JCS_GRAYSCALE")
        >>> # work with the data
        >>> # ...
        >>> im.write_spatial(grayscale)
        """
        # choose destination
        if dstfile is None:
            if self.srcfile is None:
                raise Exception("destination not specified")
            dstfile = self.srcfile
        # execute
        self._write_spatial(dstfile, data, in_color_space, dct_method, samp_factor, qt, smoothing_factor, flags)
        # free buffers
        self._im_dct = None
        self._im_spatial = None

    def source_color_space(self):
        """Returns the color space of the source file."""
        return self.color_space

    def to_spatial(self, Y=None, CbCr=None, qt=None, quantized=False, **kw): #, qt=None):
        """Converts DCT representation to spatial.
        
        Performs decompression from DCT coefficients to spatial representation.
        Uses temporary file to save and read from as libjpeg does not have a direct handler.
        
        :param Y: Lumo DCT.
        :type Y: numpy.ndarray, optional
        :param CbCr: Chrominance DCT.
        :type CbCr: numpy.ndarray, optional
        :param qt: Quantization tables or quality.
        :type qt: numpy.ndarray | int, optional
        :param quantized: Indicates whether the input DCT coefficients are quantized or unquantized, False by default.
        :type quantized: bool, optional
        :return: Spatial representation of DCT coefficients
        :rtype: numpy.ndarray

        :Example:

        >>> im = jpeglib.JPEG("input.jpeg")
        >>> Y,CbCr,qt = im.read_dct()
        >>> spatial = im.to_spatial(Y, CbCr, out_color_space="JCS_RGB")
        """
        # TODO: grayscale
        if (Y is None or CbCr is None) and self._im_dct is None:
            raise RuntimeError("Call read_dct() before calling to_spatial() or specify Y and CbCr.")
        with tempfile.NamedTemporaryFile() as tmp:
            self.write_dct(tmp.name, Y=Y, CbCr=CbCr, qt=qt, quantized=quantized)
            with JPEG(tmp.name) as im:
                data = im.read_spatial(**kw)
        self._im_dct = None
        return data

    def _read_info(self):
        # get information
        _color_space = (ctypes.c_int*1)()
        self.cjpeglib.read_jpeg_info(
            srcfile = self.srcfile,
            dct_dims = self._dct_dims,
            image_dims = self._dims,
            num_components = self._num_components,
            samp_factor = self._samp_factor,
            jpeg_color_space = _color_space
        )
        # parse jpeg information
        self.channels = self._num_components[0] # number of color components (can be assumed 3?)
        self.dct_channels = self._num_components[0]
        self.dct_shape = np.array([self._dct_dims[i] for i in range(6)], int)\
            .reshape(self.dct_channels, 2)
        self.shape = np.array([self._dims[0], self._dims[1]])
        self.color_space = [k for k,v in self.J_COLOR_SPACE.items() if v[0] == _color_space[0]][0]

    def _allocate(self):
        self._im_spatial,self._im_colormap = self._allocate_spatial()
        self._im_dct = self._allocate_dct()

    def _initialize_info(self, data=None, Y=None, CbCr=None):
        if data is not None:
            self.dct_channels = self.channels = data.shape[2]
            self.shape = np.array(data.shape[:2])
        else:
            self.dct_channels = self.channels = 3
            self.shape = np.array([0,0])
        if Y is not None:
            self._dct_dims[0] = Y.shape[1]
            self._dct_dims[1] = Y.shape[2]
            if CbCr is not None:
                self._dct_dims[2] = CbCr.shape[1]
                self._dct_dims[3] = CbCr.shape[2]
                self._dct_dims[4] = CbCr.shape[1]
                self._dct_dims[5] = CbCr.shape[2]
                self.shape = np.array([self._dct_dims[1], self._dct_dims[0]]) * 8
                self.dct_channels = 3
            else:
                self.shape = np.array([self._dct_dims[1], self._dct_dims[0]]) * 8
                self.dct_channels = 1
            self.channels = self.dct_channels
        else:
            for i in range(6):
                self._dct_dims[i] = math.ceil((data.shape[i % 2] / 8))
                #if self._samp_factor[i // 2][i % 2] != 0:
                #    self._dct_dims[i] = math.ceil(self._dct_dims[i])# / self._samp_factor[i // 2][i % 2])

        self.dct_shape = np.array([self._dct_dims[i] for i in range(2*self.dct_channels)], int)\
            .reshape(self.dct_channels, 2)
        self._dims = (ctypes.c_int * 3)(self.shape[1], self.shape[0], self.channels)

        self._allocate()
    
    def _read_dct(self, srcfile):
        """"""
        # check source file
        if srcfile is None: raise IOError("source file not given")
        try: self.dct_shape
        except: raise IOError("source file does not exist")
        # allocate
        if self._im_dct is None:
            self._im_dct = self._allocate_dct()
        
        # reading
        self.cjpeglib.read_jpeg_dct(
            srcfile = srcfile,
            dct     = self._im_dct,
            qt      = self._im_qt
        )
        
        # align qt
        qt = np.ctypeslib.as_array(self._im_qt)
        qt = qt.reshape((*qt.shape[:-1],8,8))
        # align lumo
        Y = np.ctypeslib.as_array(self._im_dct[:1])
        Y = Y.reshape((*Y.shape[:-1],8,8))
        # align chroma
        CbCr = np.ctypeslib.as_array(self._im_dct[1:])
        CbCr = CbCr[:,:self.dct_shape[1][0],:self.dct_shape[1][1]]
        CbCr = CbCr.reshape((*CbCr.shape[:-1],8,8))
        # finish
        return Y,CbCr,qt


    def _write_dct(self, dstfile, Y, CbCr=None, quality=None, quantized=False, in_color_space=None, samp_factor=None):
        """"""
        # initialize default
        #if self.srcfile is None: # if no srcfile, internals need to be initialized
        self._initialize_info(Y=Y, CbCr=CbCr)
        samp_factor = self._parse_samp_factor(samp_factor, Y, CbCr)
        # allocate DCT
        if self._im_dct is None: # could have been previously freed
            self._im_dct = self._allocate_dct()
        # quantization table and quality
        if quantized and quality is None:
            qt = np.ctypeslib.as_array(self._im_qt)
        qtable,quality = self._parse_quality(quality)
        if quantized:
            if qtable is not None:
                qt = np.ctypeslib.as_array(qtable)
            elif self._im_qt is not None:
                qt = np.ctypeslib.as_array(self._im_qt)
            else:
                raise IOError("can't dequantize with quantization table not given")
        
        # align lumo
        Y = Y.reshape((*Y.shape[:-2],64))
        if quantized:
            Y = (Y / qt[0]).astype(np.int16)
        self._im_dct[0] = np.ctypeslib.as_ctypes(Y[0])
        # align chroma
        if CbCr is not None: # not given e.g. for grayscale
            CbCr = CbCr.reshape((*CbCr.shape[:-2],64))
            if quantized:
                CbCr = (CbCr / qt[1]).astype(np.int16)
            _CbCr = np.zeros((2, self.dct_shape[0][0], self.dct_shape[0][1], 64), np.short)
            _CbCr[:,:int(self.dct_shape[1][0]),:int(self.dct_shape[1][1])] = CbCr
            self._im_dct[1:] = np.ctypeslib.as_ctypes(_CbCr)
        
        # colorspace and channels
        if in_color_space is None:
            if CbCr is None: in_color_space = 'JCS_GRAYSCALE'
            else: in_color_space = 'JCS_RGB'
            #in_color_space = self.color_space
        in_color_space,channels = self.J_COLOR_SPACE[in_color_space]
        
        
        # write
        #print("write_jpeg_dct:", self._dims[:], in_color_space, self.channels, [samp_factor[i] for i in range(3)])
        self.cjpeglib.write_jpeg_dct(
            srcfile        = self.srcfile,
            dstfile        = dstfile,
            dct            = self._im_dct,
            image_dims     = self._dims,
            in_color_space = in_color_space,
            in_components  = self.channels,
            samp_factor    = samp_factor,
            qt             = qtable,
            quality        = quality
        )


    def _read_spatial(self, srcfile, out_color_space, dither_mode, dct_method, colormap, flags):
        """"""
        # check source file
        if srcfile is None: raise IOError("source file not given")
        
        # color space and channels
        if out_color_space is None and self.color_space is not None:
            out_color_space = self.color_space
        color_space,channels = self.J_COLOR_SPACE[out_color_space]
        # dither mode and dct method
        dither_mode = self.J_DITHER_MODE[dither_mode]
        dct_method = self.J_DCT_METHOD[dct_method]
        
        if any(i in flags for i in ['QUANTIZE_COLORS','+QUANTIZE_COLORS','-QUANTIZE_COLORS']):
            raise NotImplementedError("color quantization not fully implemented")
        in_cmap = None
        if colormap is not None:
            raise NotImplementedError("color quantization not fully implemented")
            #if 'QUANTIZE_COLORS' not in flags and '+QUANTIZE_COLORS' not in flags:
            #    flags.append('+QUANTIZE_COLORS')
            #in_cmap = np.ascontiguousarray(colormap)
            #in_cmap = np.ctypeslib.as_ctypes(in_cmap.astype(np.ubyte))
        
        # allocate buffers
        if self._im_spatial is None or channels != self.channels:
            if channels > 0:
                self.channels = channels
            self._im_spatial,self._im_colormap = self._allocate_spatial()
        else: self.channels = channels # allocated correctly

        # call
        self.cjpeglib.read_jpeg_spatial(
            srcfile         = srcfile,
            rgb             = self._im_spatial,
            colormap        = self._im_colormap,
            in_colormap     = in_cmap,
            out_color_space = color_space,
            dither_mode     = dither_mode,
            dct_method      = dct_method,
            flags           = flags
        )
        
        # align rgb
        data = np.ctypeslib.as_array(self._im_spatial).astype(np.ubyte)
        
        # if 'QUANTIZE_COLORS' in flags and '+QUANTIZE_COLORS' in flags:
        #     data = data.reshape(self.channels,data.shape[2],-1)[0]
        #     # parse colormap
        #     colormap = np.ctypeslib.as_array(self._im_colormap)
        #     colormap = colormap.reshape(colormap.shape[1], self.channels)
        #     # index to color
        #     data = np.array([[colormap[i] for i in row] for row in np.array(data)])
        # else:
        #     #print(data.shape, self.channels)
        #     data = data.reshape(data.shape[2],-1,self.channels)
        # quantization
        #if 'QUANTIZE_COLORS' in flags:
        #    data = data[:,:,0]
        data = data.reshape(data.shape[2],-1,self.channels)
        # finish
        return data
    
    def _write_spatial(self, dstfile, data, in_color_space, dct_method, samp_factor, quality, smoothing_factor, flags):
        """"""
        # initialize default
        #if self.srcfile is None:
        #    self._initialize_info(data=data)
        self._initialize_info(data=data)
        self._samp_factor = self._parse_samp_factor(samp_factor)
        
        # parameters
        if in_color_space is None:
            if data.shape[2] == 1:
                in_color_space = 'JCS_GRAYSCALE'
            elif data.shape[2] == 3:
                in_color_space = 'JCS_RGB'
            else:
                raise Exception('unknown colorspace')
        #in_color_space = in_color_space if in_color_space is not None else self.color_space
        color_space,channels = self.J_COLOR_SPACE[in_color_space]
        dct_method = self.J_DCT_METHOD[dct_method]
        qt,quality = self._parse_quality(quality)
        
        # buffer
        if data is not None:
            data = data.reshape(self.channels,data.shape[1],-1)
            self._im_spatial = np.ctypeslib.as_ctypes(data)
            self._dims = (ctypes.c_int * 3)(data.shape[1], data.shape[2], data.shape[0])
        elif self._im_spatial is None:
            warnings.warn('Writing unsuccessful, call read_spatial() before calling write_spatial() or specify data parameter.', RuntimeWarning)
            return
        # call
        self.cjpeglib.write_jpeg_spatial(
            dstfile          = dstfile,
            rgb              = self._im_spatial,
            image_dims       = self._dims,
            in_color_space   = color_space,
            in_components    = self.channels,
            dct_method       = dct_method,
            samp_factor      = self._samp_factor,
            quality          = quality,
            qt               = qt,
            smoothing_factor = smoothing_factor,
            flags            = flags
        )

    def print_params(self): self.cjpeglib.print_jpeg_params(self.srcfile)
    def __enter__(self):
        """Method for using ``with`` statement together with :class:`JPEG`.
        
        :Example:
        
        >>> with jpeglib.JPEG("input.jpeg") as im:
        >>>     Y,CbCr,qt = im.read_dct()
        """
        return self
    def __exit__(self, exception_type, exception_val, trace):
        """Method for using ``with`` statement together with :class:`JPEG`."""
        self.close()
    def close(self):
        """Closes the object. Defined for interface compatibility with PIL.
        
        :Example:

        >>> im = jpeglib.JPEG("input.jpeg")
        >>> # work with im
        >>> im.close()
        """
        pass
    def _allocate_dct(self):
        return ((((ctypes.c_short * 64) 
                                  * self.dct_shape[0][1])
                                  * self.dct_shape[0][0])
                                  * self.dct_channels)()
    def _allocate_spatial(self):
        return ((((ctypes.c_ubyte * self.shape[1])
                                 * self.shape[0])
                                 * self.channels)(),
                ((ctypes.c_ubyte * 256) * self.channels)())
        
    def _parse_samp_factor(self, samp_factor=None, Y=None, CbCr=None):
        # sampling factor not given
        if samp_factor is None:
            # deduce sampling from shape of Y and CbCr
            if Y is not None and CbCr is not None:
                hs = [Y.shape[1], CbCr.shape[1], CbCr.shape[1]]
                ws = [Y.shape[2], CbCr.shape[2], CbCr.shape[2]]
                hmax,wmax = max(hs),max(ws)
                hs,ws = [h/hmax for h in hs],[w/wmax for w in ws]
                hmin,wmin = min(hs),min(ws)
                hs,ws = [round(h/hmin) for h in hs],[round(w/wmin) for w in ws]
                samp_factor = [[h,w] for h,w in zip(hs,ws)]
        # default samp factor
        #if samp_factor is None:
        #    samp_factor = [[2,2],[1,1],[1,1]]
        # process given sampling factor
        if samp_factor is not None:
            assert(len(samp_factor) == 3)
            # J:a:b notation
            if [len(s) for s in samp_factor] == [1,1,1]:
                samp_factor = Jab_to_factors(samp_factor)
        # check factors are valid
        if samp_factor is not None:
            samp_factor = np.array(samp_factor, dtype=np.int32)
            if np.sum(np.prod(samp_factor, axis=1)) > 10:
                raise Exception("linear combination of samp_factor greater than 10")
            if np.any(~np.isin(samp_factor,[1,2,3,4])):
                raise Exception("factors not in 1,2,3,4")
            # convert to ctypes
            self._samp_factor = np.ctypeslib.as_ctypes(samp_factor)
        return self._samp_factor
    
    def _parse_quality(self, quality):
        if quality is None: # not specified
            qt = self._im_qt
        else:
            # numeric quality
            try:
                quality = int(quality)
                qt = None
            # quantization table
            except:
                qt = np.array(quality, dtype=np.uint16)#.reshape(2,64)
                qt = self._im_qt = np.ctypeslib.as_ctypes(qt)
                quality = -1
        return qt,quality




__all__ = ["JPEG"]